"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
from typing import Tuple, List, Union

import tomli
from rdflib.plugins.parsers.notation3 import BadSyntax

from . import logger_setup
from dataclasses import dataclass

from pyparsing import ParseException
from pathlib import Path
from requests import ConnectionError, ConnectTimeout, HTTPError, RequestException

from rdflib import Graph, URIRef, RDF, XSD, SH, Literal

from rdflib.compare import isomorphic, graph_diff
import pandas

from .namespace import MUST, TRIPLESTORE
import requests
import json
from pandas import DataFrame

from .spec_component import TableThenSpec, parse_spec_component, WhenSpec, ThenSpec
from .utils import is_json, get_mustrd_root
from colorama import Fore, Style
from tabulate import tabulate
from collections import defaultdict
from pyshacl import validate
import logging
from http.client import HTTPConnection
from .steprunner import upload_given, run_when
from multimethods import MultiMethod
import traceback

log = logging.getLogger(__name__)

requests.packages.urllib3.disable_warnings()
requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def debug_requests_on():
    '''Switches on logging of the requests module.'''
    HTTPConnection.debuglevel = 1

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def debug_requests_off():
    '''Switches off logging of the requests module, might be some side-effects'''
    HTTPConnection.debuglevel = 0

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.WARNING)
    root_logger.handlers = []
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.WARNING)
    requests_log.propagate = False


debug_requests_off()


@dataclass(frozen=True)
class Specification:
    spec_uri: URIRef
    triple_store: dict
    given: Graph
    when: WhenSpec
    then: ThenSpec
    spec_file_name: str = "default.mustrd.ttl"
    spec_source_file: Path = Path("default.mustrd.ttl")


@dataclass
class GraphComparison:
    in_expected_not_in_actual: Graph
    in_actual_not_in_expected: Graph
    in_both: Graph


@dataclass
class SpecResult:
    spec_uri: URIRef
    triple_store: URIRef


@dataclass
class SpecPassed(SpecResult):
    pass


@dataclass()
class SpecPassedWithWarning(SpecResult):
    warning: str


@dataclass
class SelectSpecFailure(SpecResult):
    table_comparison: pandas.DataFrame
    message: str


@dataclass
class ConstructSpecFailure(SpecResult):
    graph_comparison: GraphComparison


@dataclass
class UpdateSpecFailure(SpecResult):
    graph_comparison: GraphComparison


@dataclass
class SparqlParseFailure(SpecResult):
    exception: ParseException


@dataclass
class SparqlExecutionError(SpecResult):
    exception: Exception


@dataclass
class TripleStoreConnectionError(SpecResult):
    exception: ConnectionError


@dataclass
class SpecSkipped(SpecResult):
    message: str
    spec_file_name: str = "default.mustrd.ttl"
    spec_source_file: Path = Path("default.mustrd.ttl")


@dataclass
class SparqlAction:
    query: str


@dataclass
class SelectSparqlQuery(SparqlAction):
    pass


@dataclass
class ConstructSparqlQuery(SparqlAction):
    pass


@dataclass
class UpdateSparqlQuery(SparqlAction):
    pass


# https://github.com/Semantic-partners/mustrd/issues/19
# Validate the specs found in spec_path
def validate_specs(run_config: dict,
                   triple_stores: List,
                   shacl_graph: Graph,
                   ont_graph: Graph,
                   file_name: str = "*",
                   selected_test_files: List[str] = [])\
        -> Tuple[List, Graph, List]:
    spec_graph = Graph()
    subject_uris = set()
    focus_uris = set()
    invalid_specs = []
    ttl_files = []

    if not selected_test_files:
        ttl_files = list(run_config['spec_path'].glob(
            f'**/{file_name}.mustrd.ttl'))
        log.info(
            f"Found {len(ttl_files)} {file_name}.mustrd.ttl files in {run_config['spec_path']}")
    else:
        ttl_files = selected_test_files

    log.info(f"Using {ttl_files} for test source")
    ttl_files.sort()

    # For each spec file found in spec_path
    for file in ttl_files:
        # file = file.resolve()
        error_messages = []

        log.info(f"Parse: {file}")
        # Parse spec file and add error message if not conform to RDF standard
        try:
            file_graph = Graph().parse(file)
        except BadSyntax as e:
            template = "An exception of type {0} occurred when trying to parse a spec file. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            log.error(message)
            error_messages += [f"Could not extract spec from {file} due to exception of type "
                               f"{type(e).__name__} when parsing file"]
            continue

        # run shacl validation
        conforms, results_graph, results_text = validate(file_graph,
                                                         shacl_graph=shacl_graph,
                                                         ont_graph=ont_graph,
                                                         inference='none',
                                                         abort_on_first=False,
                                                         allow_infos=False,
                                                         allow_warnings=False,
                                                         meta_shacl=False,
                                                         advanced=True,
                                                         js=False,
                                                         debug=False)
        if str(file.name).endswith("_duplicate"):
            log.debug(f"Validation of {file.name} against SHACL shapes: {conforms}")
            log.debug(f"{results_graph.serialize(format='turtle')}")
        # log.debug(f"SHACL validation results: {results_text}")
        # Add error message if not conform to spec shapes
        if not conforms:
            for msg in results_graph.objects(predicate=SH.resultMessage):
                log.warning(f"{file_graph}")
                log.warning(f"{msg} File: {file.name}")
                error_messages += [f"{msg} File: {file.name}"]

        # collect a list of uris of the tests in focus
        # If focus is found, only the spec in the focus will be executed
        for focus_uri in file_graph.subjects(predicate=MUST.focus, object=Literal("true", datatype=XSD.boolean)):
            if focus_uri in focus_uris:
                focus_uri = URIRef(str(focus_uri) + "_DUPLICATE")
            focus_uris.add(focus_uri)

        add_spec_validation(file_graph, subject_uris, file,
                            triple_stores, error_messages, invalid_specs, spec_graph)

    valid_spec_uris = list(spec_graph.subjects(RDF.type, MUST.TestSpec))

    if focus_uris:
        invalid_focus_specs = get_invalid_focus_spec(focus_uris, invalid_specs)
        return focus_uris, spec_graph, invalid_focus_specs
    else:
        log.info(f"Collected {len(valid_spec_uris)} valid test spec(s)")
        return valid_spec_uris, spec_graph, invalid_specs


def get_invalid_focus_spec(focus_uris: set, invalid_specs: list):
    invalid_focus_specs = []
    for spec in invalid_specs:
        if spec.spec_uri in focus_uris:
            invalid_focus_specs += [spec]
            focus_uris.remove(spec.spec_uri)
    log.info(f"Collected {len(focus_uris)} focus test spec(s)")
    return invalid_focus_specs


# Detect duplicate,
# If no error: associate the spec configuration and the file where this conf is stored
# If error, aggregate the messages and mark spec as skipped
def add_spec_validation(file_graph: Graph, subject_uris: set, file: Path, triple_stores: List,
                        error_messages: list, invalid_specs: list, spec_graph: Graph):

    for subject_uri in file_graph.subjects(RDF.type, MUST.TestSpec):
        # Always add file name and source file to the graph for error reporting
        file_graph.add([subject_uri, MUST.specSourceFile, Literal(str(file))])
        file_graph.add([subject_uri, MUST.specFileName, Literal(file.name)])

        # If we already collected a URI, then we tag it as duplicate and it won't be executed
        if subject_uri in subject_uris:
            log.warning(
                f"Duplicate subject URI found: {file.name} {subject_uri}. File will not be parsed.")
            error_messages += [f"Duplicate subject URI found in {file.name}."]
            subject_uri = URIRef(str(subject_uri) + "_DUPLICATE")
        if len(error_messages) == 0:
            subject_uris.add(subject_uri)
            this_spec_graph = Graph()
            this_spec_graph.parse(file)
            spec_uris_in_this_file = list(
                this_spec_graph.subjects(RDF.type, MUST.TestSpec))
            for spec in spec_uris_in_this_file:
                this_spec_graph.add([spec, MUST.specSourceFile, Literal(file)])
                this_spec_graph.add(
                    [spec, MUST.specFileName, Literal(file.name)])
            spec_graph += this_spec_graph
        else:
            error_messages.sort()
            error_message = "\n".join(msg for msg in error_messages)
            invalid_specs += [SpecSkipped(subject_uri, triple_store["type"], error_message, file.name, file)
                              for triple_store in triple_stores]


def get_specs(spec_uris: List[URIRef], spec_graph: Graph, triple_stores: List[dict],
              run_config: dict):
    specs = []
    skipped_results = []
    try:
        for triple_store in triple_stores:
            if "error" in triple_store:
                log.error(
                    f"{triple_store['error']}. No specs run for this triple store.")
                skipped_results += [SpecSkipped(spec_uri, triple_store['type'], triple_store['error'],
                                                get_spec_file(spec_uri, spec_graph)) for spec_uri in
                                    spec_uris]
            else:
                for spec_uri in spec_uris:
                    try:
                        specs += [get_spec(spec_uri, spec_graph,
                                           run_config, triple_store)]
                    except (ValueError, FileNotFoundError, ConnectionError) as e:
                        # Try to get file name/path from the graph, but fallback to "unknown"
                        file_name = spec_graph.value(subject=spec_uri, predicate=MUST.specFileName) or "unknown"
                        file_path = spec_graph.value(subject=spec_uri, predicate=MUST.specSourceFile) or "unknown"
                        skipped_results += [SpecSkipped(spec_uri, triple_store['type'],
                                                        str(e), str(file_name), Path(file_path))]

    except (BadSyntax, FileNotFoundError) as e:
        template = "An exception of type {0} occurred when trying to parse the triple store configuration file. " \
                   "Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.error(message)
        log.error("No specifications will be run.")

    log.info(f"Extracted {len(specs)} specifications that will be run")
    return specs, skipped_results


def run_specs(specs) -> List[SpecResult]:
    results = []
    # https://github.com/Semantic-partners/mustrd/issues/115
    for specification in specs:
        results.append(run_spec(specification))
    return results


def get_spec_file(spec_uri: URIRef, spec_graph: Graph):
    file_name = spec_graph.value(subject=spec_uri, predicate=MUST.specFileName)
    if file_name:
        return str(file_name)
    # fallback: try to get from MUST.specSourceFile
    file_path = spec_graph.value(subject=spec_uri, predicate=MUST.specSourceFile)
    if file_path:
        return str(Path(file_path).name)
    return "default.mustrd.ttl"


def get_spec(spec_uri: URIRef, spec_graph: Graph, run_config: dict, mustrd_triple_store: dict = None) -> Specification:
    try:
        if not mustrd_triple_store:
            mustrd_triple_store = {"type": TRIPLESTORE.RdfLib}
        components = []
        for predicate in MUST.given, MUST.when, MUST.then:
            components.append(parse_spec_component(subject=spec_uri,
                                                   predicate=predicate,
                                                   spec_graph=spec_graph,
                                                   run_config=run_config,
                                                   mustrd_triple_store=mustrd_triple_store))

        spec_file_name = get_spec_file(spec_uri, spec_graph)
        spec_file_path = Path(spec_graph.value(
            subject=spec_uri, predicate=MUST.specSourceFile, default=Path("default.mustrd.ttl")))
        # https://github.com/Semantic-partners/mustrd/issues/92
        return Specification(spec_uri, mustrd_triple_store,
                             components[0].value, components[1], components[2], spec_file_name, spec_file_path)

    except (ValueError, FileNotFoundError) as e:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.exception(message)
        raise
    except ConnectionError as e:
        log.error(e)
        raise


def check_result(spec: Specification, result: Union[str, Graph]):
    
    log.debug(
        f"check_result {spec.spec_uri=}, {spec.triple_store=}, {result=} {type(spec.then)}")
    if isinstance(spec.then, TableThenSpec):
        log.debug("table_comparison")
        return table_comparison(result, spec)
    else:
        graph_compare = graph_comparison(spec.then.value, result)
        if isomorphic(result, spec.then.value):
            log.debug(f"isomorphic {spec}")
            log.debug(f"{spec.spec_uri}")
            log.debug(f"{spec.triple_store}")
            ret = SpecPassed(spec.spec_uri, spec.triple_store["type"])

            return ret
        else:
            log.debug("not isomorphic")
            if spec.when[0].queryType == MUST.ConstructSparql:
                log.debug("ConstructSpecFailure")
                return ConstructSpecFailure(spec.spec_uri, spec.triple_store["type"], graph_compare)
            else:
                log.debug("UpdateSpecFailure")
                return UpdateSpecFailure(spec.spec_uri, spec.triple_store["type"], graph_compare)


def run_spec(spec: Specification) -> SpecResult:
    spec_uri = spec.spec_uri
    triple_store = spec.triple_store

    if not isinstance(spec, Specification):
        log.warning(f"check_result called with non-Specification: {type(spec)}")
        return spec
        # return SpecSkipped(getattr(spec, 'spec_uri', None), getattr(spec, 'triple_store', {}), "Spec is not a valid Specification instance")
    
    log.debug(
        f"run_spec {spec=}")
    log.debug(
        f"run_when {spec_uri=}, {triple_store=}, {spec.given=}, {spec.when=}, {spec.then=}")
    if spec.given:
        given_as_turtle = spec.given.serialize(format="turtle")
        log.debug(f"{given_as_turtle}")
        upload_given(triple_store, spec.given)
    else:
        if triple_store['type'] == TRIPLESTORE.RdfLib:
            return SpecSkipped(spec_uri, triple_store['type'], "Unable to run Inherited State tests on Rdflib")
    try:
        for when in spec.when:
            log.info(
                f"Running {when.queryType} spec {spec_uri} on {triple_store['type']}")
            try:
                result = run_when(spec_uri, triple_store, when)
                log.info(
                    f"run {when.queryType} spec {spec_uri} on {triple_store['type']} {result=}")
            except ParseException as e:
                log.error(
                    f"parseException {e}")
                return SparqlParseFailure(spec_uri, triple_store["type"], e)
            except NotImplementedError as ex:
                log.error(f"NotImplementedError {ex}")
                raise ex
                # return SpecSkipped(spec_uri, triple_store["type"], ex.args[0])
        return check_result(spec, result)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout, OSError) as e:
        # close_connection = False
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(e).__name__, e.args)
        log.error(message)
        return TripleStoreConnectionError(spec_uri, triple_store["type"], message)
    except (TypeError, RequestException) as e:
        log.error(f"{type(e)} {e}")
        return SparqlExecutionError(spec_uri, triple_store["type"], e)
    except Exception as e:
        log.error(f"Unexpected error {e}")
        return RuntimeError(spec_uri, triple_store["type"], f"{type(e).__name__}: {e}")
    # https://github.com/Semantic-partners/mustrd/issues/78
    # finally:
    #     if type(mustrd_triple_store) == MustrdAnzo and close_connection:
    #         mustrd_triple_store.clear_graph()


def get_triple_store_graph(triple_store_graph_path: Path, secrets: str):
    if secrets:
        return Graph().parse(triple_store_graph_path).parse(data=secrets)
    else:
        secret_path = triple_store_graph_path.parent / Path(triple_store_graph_path.stem +
                                                            "_secrets" + triple_store_graph_path.suffix)
        return Graph().parse(triple_store_graph_path).parse(secret_path)


# Parse and validate triple store configuration
def get_triple_stores(triple_store_graph: Graph) -> list[dict]:
    triple_stores = []
    shacl_graph = Graph().parse(
        Path(os.path.join(get_mustrd_root(), "model/triplestoreshapes.ttl")))
    ont_graph = Graph().parse(
        Path(os.path.join(get_mustrd_root(), "model/triplestoreOntology.ttl")))
    # SHACL validation of triple store configuration
    conforms, results_graph, results_text = validate(
        data_graph=triple_store_graph,
        shacl_graph=shacl_graph,
        ont_graph=ont_graph,
        advanced=True,
        inference='none'
    )
    if not conforms:
        raise ValueError(f"Triple store configuration not conform to the shapes. SHACL report: {results_text}",
                         results_graph)
    for triple_store_config, rdf_type, triple_store_type in triple_store_graph.triples((None, RDF.type, None)):
        triple_store = {}
        triple_store["type"] = triple_store_type
        triple_store["uri"] = triple_store_config
        # Anzo graph via anzo
        if triple_store_type == TRIPLESTORE.Anzo:
            get_anzo_configuration(
                triple_store, triple_store_graph, triple_store_config)
        # GraphDB
        elif triple_store_type == TRIPLESTORE.GraphDb:
            get_graphDB_configuration(
                triple_store, triple_store_graph, triple_store_config)

        elif triple_store_type != TRIPLESTORE.RdfLib:
            triple_store["error"] = f"Triple store not implemented: {triple_store_type}"

        triple_stores.append(triple_store)
    return triple_stores


def get_anzo_configuration(triple_store: dict, triple_store_graph: Graph, triple_store_config: URIRef):
    triple_store["url"] = triple_store_graph.value(
        subject=triple_store_config, predicate=TRIPLESTORE.url)
    triple_store["port"] = triple_store_graph.value(
        subject=triple_store_config, predicate=TRIPLESTORE.port)
    try:
        triple_store["username"] = str(triple_store_graph.value(subject=triple_store_config,
                                                                predicate=TRIPLESTORE.username))
        triple_store["password"] = str(triple_store_graph.value(subject=triple_store_config,
                                                                predicate=TRIPLESTORE.password))
    except (FileNotFoundError, ValueError) as e:
        triple_store["error"] = e
    triple_store["gqe_uri"] = triple_store_graph.value(subject=triple_store_config,
                                                       predicate=TRIPLESTORE.gqeURI)
    triple_store["input_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                           predicate=TRIPLESTORE.inputGraph)
    triple_store["output_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                            predicate=TRIPLESTORE.outputGraph)
    try:
        check_triple_store_params(
            triple_store, ["url", "port", "username", "password", "input_graph"])
    except ValueError as e:
        triple_store["error"] = e


def get_graphDB_configuration(triple_store: dict, triple_store_graph: Graph, triple_store_config: URIRef):
    triple_store["url"] = triple_store_graph.value(
        subject=triple_store_config, predicate=TRIPLESTORE.url)
    triple_store["port"] = triple_store_graph.value(
        subject=triple_store_config, predicate=TRIPLESTORE.port)
    try:
        triple_store["username"] = str(triple_store_graph.value(subject=triple_store_config,
                                                                predicate=TRIPLESTORE.username))
        triple_store["password"] = str(triple_store_graph.value(subject=triple_store_config,
                                                                predicate=TRIPLESTORE.password))
    except (FileNotFoundError, ValueError) as e:
        log.error(f"Credential retrieval failed {e}")
        triple_store["error"] = e
    triple_store["repository"] = triple_store_graph.value(subject=triple_store_config,
                                                          predicate=TRIPLESTORE.repository)
    triple_store["input_graph"] = triple_store_graph.value(subject=triple_store_config,
                                                           predicate=TRIPLESTORE.inputGraph)
    try:
        check_triple_store_params(triple_store, ["url", "repository"])
    except ValueError as e:
        triple_store["error"] = e


def check_triple_store_params(triple_store: dict, required_params: List[str]):
    missing_params = [
        param for param in required_params if triple_store.get(param) is None]
    if missing_params:
        raise ValueError(f"Cannot establish connection to {triple_store['type']}. "
                         f"Missing required parameter(s): {', '.join(missing_params)}.")


def get_credential_from_file(triple_store_name: URIRef, credential: str, config_path: Literal) -> str:
    log.info(
        f"get_credential_from_file {triple_store_name}, {credential}, {config_path}")
    if not config_path:
        raise ValueError(f"Cannot establish connection defined in {triple_store_name}. "
                         f"Missing required parameter: {credential}.")
    path = Path(config_path)
    log.info(f"get_credential_from_file {path}")

    if not os.path.isfile(path):
        log.error(f"couldn't find {path}")
        raise FileNotFoundError(f"Credentials config file not found: {path}")
    try:
        with open(path, "rb") as f:
            config = tomli.load(f)
    except tomli.TOMLDecodeError as e:
        log.error(f"config error {path} {e}")
        raise ValueError(f"Error reading credentials config file: {e}")
    return config[str(triple_store_name)][credential]


# Convert sparql json query results as defined in https://www.w3.org/TR/rdf-sparql-json-res/
def json_results_to_panda_dataframe(result: str) -> pandas.DataFrame:
    json_result = json.loads(result)
    frames = DataFrame()
    for binding in json_result["results"]["bindings"]:
        columns = []
        values = []
        for key in binding:
            value_object = binding[key]
            columns.append(key)
            values.append(str(value_object["value"]))
            columns.append(key + "_datatype")
            if "type" in value_object and value_object["type"] == "literal":
                literal_type = str(XSD.string)
                if "datatype" in value_object:
                    literal_type = value_object["datatype"]
                values.append(literal_type)
            else:
                values.append(str(XSD.anyURI))

        frames = pandas.concat(objs=[frames, pandas.DataFrame(
            [values], columns=columns)], ignore_index=True)
        frames.fillna('', inplace=True)

        if frames.size == 0:
            frames = pandas.DataFrame()
    return frames


def table_comparison(result: str, spec: Specification) -> SpecResult:
    warning = None
    order_list = ["order by ?", "order by desc", "order by asc"]
    ordered_result = any(
        pattern in spec.when[0].value.lower() for pattern in order_list)

    # If sparql query doesn't contain order by clause, but order is define in then spec:
    # Then ignore order in then spec and print a warning
    if not ordered_result and spec.then.ordered:
        warning = f"sh:order in {spec.spec_uri} is ignored, no ORDER BY in query"
        log.warning(warning)

    # If sparql query contains an order by clause and then spec is not order:
    # Spec is inconsistent
    if ordered_result and not spec.then.ordered:
        message = "Actual result is ordered, must:then must contain sh:order on every row."
        return SelectSpecFailure(spec.spec_uri, spec.triple_store["type"], None, message)

    # Convert results to dataframe
    if is_json(result):
        df = json_results_to_panda_dataframe(result)
    else:
        return SelectSpecFailure(spec.spec_uri, spec.triple_store["type"], None, "Sparql result is not in JSON")

    # Compare result with expected
    df_diff, message = compare_table_results(df, spec)

    if df_diff.empty:
        if warning:
            return SpecPassedWithWarning(spec.spec_uri, spec.triple_store["type"], warning)
        else:
            return SpecPassed(spec.spec_uri, spec.triple_store["type"])
    else:
        log.error("\n" + df_diff.to_markdown())
        log.error(message)
        return SelectSpecFailure(spec.spec_uri, spec.triple_store["type"], df_diff, message)


def compare_table_results_dispatch(resultDf: DataFrame, spec: Specification):
    return not resultDf.empty, not spec.then.value.empty


compare_table_results = MultiMethod(
    "compare_table_results", compare_table_results_dispatch)


# Scenario 1: expected a result and got a result
@compare_table_results.method((True, True))
def _compare_results(resultDf: DataFrame, spec: Specification):
    columns = list(resultDf.columns)
    sorted_columns = sorted(columns)
    then = spec.then.value
    sorted_then_cols = sorted(list(then))
    order_list = ["order by ?", "order by desc", "order by asc"]
    ordered_result = any(
        pattern in spec.when[0].value.lower() for pattern in order_list)

    if not ordered_result:
        resultDf.sort_values(by=list(resultDf.columns)[::2], inplace=True)
        resultDf.reset_index(inplace=True, drop=True)

    if len(columns) == len(then.columns):
        if sorted_columns == sorted_then_cols:
            then = then[columns]
            if not ordered_result:
                then.sort_values(by=columns[::2], inplace=True)
                then.reset_index(drop=True, inplace=True)
            if resultDf.shape == then.shape and (resultDf.columns == then.columns).all():
                df_diff = then.compare(
                    resultDf, result_names=("expected", "actual"))
            else:
                df_diff = construct_df_diff(resultDf, then)
        else:
            then = then[sorted_then_cols]
            resultDf = resultDf[sorted_columns]
            df_diff = construct_df_diff(resultDf, then)
    else:
        then = then[sorted_then_cols]
        resultDf = resultDf[sorted_columns]
        df_diff = construct_df_diff(resultDf, then)

    message = build_summary_message(then.shape[0], round(
        then.shape[1] / 2), resultDf.shape[0], round(resultDf.shape[1] / 2))
    return df_diff, message


# Scenario 2: expected no result but got a result
@compare_table_results.method((True, False))
def _unexpected_results(resultDf: DataFrame, spec: Specification):
    empty_then = create_empty_dataframe_with_columns(resultDf)
    df_diff = empty_then.compare(resultDf, result_names=("expected", "actual"))

    return df_diff, build_summary_message(0, 0, resultDf.shape[0], round(resultDf.shape[1] / 2))


# Scenario 3: expected a result, but got an empty result
@compare_table_results.method((False, True))
def _missing_results(resultDf: DataFrame, spec: Specification):
    then = spec.then.value
    then = then[sorted(list(then))]
    df = create_empty_dataframe_with_columns(then)
    df_diff = then.compare(df, result_names=("expected", "actual"))

    return df_diff, build_summary_message(then.shape[0], round(then.shape[1] / 2), 0, 0)


# Scenario 4: expected no result, got no result
@compare_table_results.method((False, False))
def _no_results(resultDf: DataFrame, spec: Specification):
    df = pandas.DataFrame()
    df_diff = spec.then.value.compare(df, result_names=("expected", "actual"))

    return df_diff, build_summary_message(0, 0, 0, 0)


def build_summary_message(expected_rows, expected_columns, got_rows, got_columns):
    return f"Expected {expected_rows} row(s) and {expected_columns} column(s), " \
        f"got {got_rows} row(s) and {got_columns} column(s)"


def graph_comparison(expected_graph: Graph, actual_graph: Graph) -> GraphComparison:
    diff = graph_diff(expected_graph, actual_graph)
    in_both = diff[0]
    in_expected = diff[1]
    in_actual = diff[2]
    in_expected_not_in_actual = (in_expected - in_actual)
    in_actual_not_in_expected = (in_actual - in_expected)
    return GraphComparison(in_expected_not_in_actual, in_actual_not_in_expected, in_both)


def get_then_update(spec_uri: URIRef, spec_graph: Graph) -> Graph:
    then_query = f"""
    prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

    CONSTRUCT {{ ?s ?p ?o }}
    {{
        <{spec_uri}> <{MUST.then}>
            a <{MUST.StatementsDataset}> ;
            <{MUST.hasStatement}> [
                a rdf:Statement ;
                rdf:subject ?s ;
                rdf:predicate ?p ;
                rdf:object ?o ;
            ] ; ]
    }}
    """
    expected_results = spec_graph.query(then_query).graph

    return expected_results


def write_result_diff_to_log(res, info):
    if isinstance(res, UpdateSpecFailure) or isinstance(res, ConstructSpecFailure):
        info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
        info(f"{Fore.BLUE} In Expected Not In Actual:")
        info(
            res.graph_comparison.in_expected_not_in_actual.serialize(format="ttl"))
        info(f"{Fore.RED} in_actual_not_in_expected")
        info(
            res.graph_comparison.in_actual_not_in_expected.serialize(format="ttl"))
        info(f"{Fore.GREEN} in_both")
        info(res.graph_comparison.in_both.serialize(format="ttl"))

    if isinstance(res, SelectSpecFailure):
        info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
        info(res.message)
        info(res.table_comparison.to_markdown())
    if isinstance(res, SpecPassedWithWarning):
        info(
            f"{Fore.YELLOW}Passed with warning {res.spec_uri} {res.triple_store}")
        info(res.warning)
    if isinstance(res, TripleStoreConnectionError) or isinstance(res, SparqlExecutionError) or \
            isinstance(res, SparqlParseFailure):
        info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
        info(res.exception)
    if isinstance(res, SpecSkipped):
        info(f"{Fore.YELLOW}Skipped {res.spec_uri} {res.triple_store}")
        info(res.message)


def calculate_row_difference(df1: pandas.DataFrame,
                             df2: pandas.DataFrame) -> pandas.DataFrame:
    df_all = df1.merge(df2.drop_duplicates(), how='left', indicator=True)
    actual_rows = df_all[df_all['_merge'] == 'left_only']
    actual_rows = actual_rows.drop('_merge', axis=1)
    return actual_rows


def construct_df_diff(df: pandas.DataFrame,
                      then: pandas.DataFrame) -> pandas.DataFrame:
    actual_rows = calculate_row_difference(df, then)
    expected_rows = calculate_row_difference(then, df)
    actual_columns = df.columns.difference(then.columns)
    expected_columns = then.columns.difference(df.columns)

    df_diff = pandas.DataFrame()
    modified_df = df
    modified_then = then

    if actual_columns.size > 0:
        modified_then = modified_then.reindex(
            modified_then.columns.to_list() + actual_columns.to_list(), axis=1)
        modified_then[actual_columns.to_list(
        )] = modified_then[actual_columns.to_list()].fillna('')

    if expected_columns.size > 0:
        modified_df = modified_df.reindex(
            modified_df.columns.to_list() + expected_columns.to_list(), axis=1)
        modified_df[expected_columns.to_list(
        )] = modified_df[expected_columns.to_list()].fillna('')

    modified_df = modified_df.reindex(modified_then.columns, axis=1)

    if df.shape[0] != then.shape[0] and df.shape[1] != then.shape[1]:
        # take modified columns and add rows
        actual_rows = calculate_row_difference(modified_df, modified_then)
        expected_rows = calculate_row_difference(modified_then, modified_df)
        df_diff = generate_row_diff(actual_rows, expected_rows)
    elif actual_rows.shape[0] > 0 or expected_rows.shape[0] > 0:
        df_diff = generate_row_diff(actual_rows, expected_rows)
    elif actual_columns.size > 0 or expected_columns.size > 0:
        df_diff = modified_then.compare(modified_df, result_names=("expected", "actual"), keep_shape=True,
                                        keep_equal=True)
    df_diff.fillna("", inplace=True)
    return df_diff


def generate_row_diff(actual_rows: pandas.DataFrame, expected_rows: pandas.DataFrame) -> pandas.DataFrame:
    df_diff_actual_rows = pandas.DataFrame()
    df_diff_expected_rows = pandas.DataFrame()

    if actual_rows.shape[0] > 0:
        empty_actual_copy = create_empty_dataframe_with_columns(actual_rows)
        df_diff_actual_rows = empty_actual_copy.compare(
            actual_rows, result_names=("expected", "actual"))

    if expected_rows.shape[0] > 0:
        empty_expected_copy = create_empty_dataframe_with_columns(
            expected_rows)
        df_diff_expected_rows = expected_rows.compare(
            empty_expected_copy, result_names=("expected", "actual"))

    df_diff_rows = pandas.concat(
        [df_diff_actual_rows, df_diff_expected_rows], ignore_index=True)
    return df_diff_rows


def create_empty_dataframe_with_columns(df: pandas.DataFrame) -> pandas.DataFrame:
    empty_copy = pandas.DataFrame().reindex_like(df)
    empty_copy.fillna("", inplace=True)
    return empty_copy


def review_results(results: List[SpecResult], verbose: bool) -> None:
    log.info("===== Result Overview =====")
    # Init dictionaries
    status_dict = defaultdict(lambda: defaultdict(int))
    status_counts = defaultdict(lambda: defaultdict(int))
    colours = {SpecPassed: Fore.GREEN,
               SpecPassedWithWarning: Fore.YELLOW, SpecSkipped: Fore.YELLOW}
    # Populate dictionaries from results
    for result in results:
        status_counts[result.triple_store][type(result)] += 1
        status_dict[result.spec_uri][result.triple_store] = type(result)

    # Get the list of statuses and list of unique triple stores
    statuses = list(status for inner_dict in status_dict.values()
                    for status in inner_dict.values())
    triple_stores = list(set(status for inner_dict in status_dict.values()
                         for status in inner_dict.keys()))

    # Convert dictionaries to list for tabulate
    table_rows = [[spec_uri] + [
        f"""{colours.get(status_dict[spec_uri][triple_store], Fore.RED)}
        {status_dict[spec_uri][triple_store].__name__}{Style.RESET_ALL}"""
        for triple_store in triple_stores] for spec_uri in set(status_dict.keys())]

    status_rows = [[f"{colours.get(status, Fore.RED)}{status.__name__}{Style.RESET_ALL}"] +
                   [f"{colours.get(status, Fore.RED)}{status_counts[triple_store][status]}{Style.RESET_ALL}"
                    for triple_store in triple_stores] for status in set(statuses)]

    # Display tables with tabulate
    log.info(tabulate(table_rows, headers=[
             'Spec Uris / triple stores'] + triple_stores, tablefmt="pretty"))
    log.info(tabulate(status_rows, headers=[
             'Status / triple stores'] + triple_stores, tablefmt="pretty"))

    pass_count = statuses.count(SpecPassed)
    warning_count = statuses.count(SpecPassedWithWarning)
    skipped_count = statuses.count(SpecSkipped)
    fail_count = len(
        list(filter(lambda status: status not in [SpecPassed, SpecPassedWithWarning, SpecSkipped], statuses)))

    if fail_count:
        overview_colour = Fore.RED
    elif warning_count or skipped_count:
        overview_colour = Fore.YELLOW
    else:
        overview_colour = Fore.GREEN

    logger_setup.flush()
    log.info(f"{overview_colour}===== {fail_count} failures, {skipped_count} skipped, {Fore.GREEN}{pass_count} passed, "
             f"{overview_colour}{warning_count} passed with warnings =====")

    if verbose and (fail_count or warning_count or skipped_count):
        display_verbose(results)


def display_verbose(results: List[SpecResult]):
    for res in results:
        if isinstance(res, UpdateSpecFailure):
            log.info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
            log.info(f"{Fore.BLUE} In Expected Not In Actual:")
            log.info(
                res.graph_comparison.in_expected_not_in_actual.serialize(format="ttl"))
            log.info()
            log.info(f"{Fore.RED} in_actual_not_in_expected")
            log.info(
                res.graph_comparison.in_actual_not_in_expected.serialize(format="ttl"))
            log.info(f"{Fore.GREEN} in_both")
            log.info(res.graph_comparison.in_both.serialize(format="ttl"))

        if isinstance(res, SelectSpecFailure):
            log.info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
            log.info(res.message)
            log.info(res.table_comparison.to_markdown())
        if isinstance(res, ConstructSpecFailure) or isinstance(res, UpdateSpecFailure):
            log.info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
        if isinstance(res, SpecPassedWithWarning):
            log.info(
                f"{Fore.YELLOW}Passed with warning {res.spec_uri} {res.triple_store}")
            log.info(res.warning)
        if isinstance(res, TripleStoreConnectionError) or type(res, SparqlExecutionError) or \
                isinstance(res, SparqlParseFailure):
            log.info(f"{Fore.RED}Failed {res.spec_uri} {res.triple_store}")
            log.info(res.exception)
        if isinstance(res, SpecSkipped):
            log.info(f"{Fore.YELLOW}Skipped {res.spec_uri} {res.triple_store}")
            log.info(res.message)



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import urllib.parse
import requests
from rdflib import Graph, Literal
from requests import ConnectionError, HTTPError, RequestException, Response


# https://github.com/Semantic-partners/mustrd/issues/72
def manage_graphdb_response(response: Response) -> str:
    content_string = response.content.decode("utf-8")
    if response.status_code == 200:
        return content_string
    elif response.status_code == 204:
        pass
    elif response.status_code == 401:
        raise HTTPError(f"GraphDB authentication error, status code: {response.status_code}, content: {content_string}")
    elif response.status_code == 406:
        raise HTTPError(f"GraphDB  error, status code: {response.status_code}, content: {content_string}")
    else:
        raise RequestException(f"GraphDb error, status code: {response.status_code}, content: {content_string}")


def upload_given(triple_store: dict, given: Graph):
    if given:
        try:
            graph = "default"
            if triple_store['input_graph']:
                graph = urllib.parse.urlencode({'graph': triple_store['input_graph']})
            url = f"{triple_store['url']}/repositories/{triple_store['repository']}" \
                  f"/rdf-graphs/service?{graph}"
            # graph store PUT drop silently the graph or default and upload the payload
            # https://www.w3.org/TR/sparql11-http-rdf-update/#http-put
            manage_graphdb_response(requests.put(url=url,
                                                 auth=(triple_store['username'], triple_store['password']),
                                                 data=given.serialize(format="ttl"),
                                                 headers={'Content-Type': 'text/turtle'}))
        except ConnectionError:
            raise


def parse_bindings(bindings: dict = None) -> dict:
    return None if not bindings else {f"${k}": str(v.n3()) for k, v in bindings.items()}


def execute_select(triple_store: dict, when: str, bindings: dict = None) -> str:
    return post_query(triple_store, when, "application/sparql-results+json", parse_bindings(bindings))


def execute_construct(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    return Graph().parse(data=post_query(triple_store, when, "text/turtle", parse_bindings(bindings)))


def execute_update(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    post_update_query(triple_store, when, parse_bindings(bindings))
    return Graph().parse(data=post_query(triple_store, "CONSTRUCT {?s ?p ?o} where { ?s ?p ?o }", 'text/turtle'))


def post_update_query(triple_store: dict, query: str, params: dict = None) -> str:
    params = add_graph_to_params(params, triple_store["input_graph"])
    try:
        return manage_graphdb_response(requests.post(
            url=f"{triple_store['url']}/repositories/{triple_store['repository']}/statements",
            data=query,
            params=params,
            auth=(triple_store['username'], triple_store['password']),
            headers={'Content-Type': 'application/sparql-update'}))
    except (ConnectionError, OSError):
        raise


def post_query(triple_store: dict, query: str, accept: str, params: dict = None) -> str:
    headers = {
        'Content-Type': 'application/sparql-query',
        'Accept': accept
    }
    params = add_graph_to_params(params, triple_store["input_graph"])
    try:
        return manage_graphdb_response(
            requests.post(url=f"{triple_store['url']}/repositories/{triple_store['repository']}",
                          data=query,
                          params=params,
                          auth=(triple_store['username'], triple_store['password']),
                          headers=headers))
    except (ConnectionError, OSError):
        raise


def add_graph_to_params(params: dict, graph: Literal) -> dict:
    graph = graph or "http://rdf4j.org/schema/rdf4j#nil"
    if params:
        params['default-graph-uri'] = graph
        params['using-graph-uri'] = graph
        params['remove-graph-uri'] = graph
        params['insert-graph-uri'] = graph
    else:
        params = {
            'default-graph-uri': graph,
            'using-graph-uri': graph,
            'remove-graph-uri': graph,
            'insert-graph-uri': graph
        }
    return params



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json

from multimethods import MultiMethod, Default
from .namespace import MUST, TRIPLESTORE
from rdflib import Graph, URIRef
from .mustrdRdfLib import execute_select as execute_select_rdflib
from .mustrdRdfLib import execute_construct as execute_construct_rdflib
from .mustrdRdfLib import execute_update as execute_update_rdflib
from .mustrdAnzo import get_query_from_step, upload_given as upload_given_anzo
from .mustrdAnzo import execute_update as execute_update_anzo
from .mustrdAnzo import execute_construct as execute_construct_anzo
from .mustrdAnzo import execute_select as execute_select_anzo
from .mustrdGraphDb import upload_given as upload_given_graphdb
from .mustrdGraphDb import execute_update as execute_update_graphdb
from .mustrdGraphDb import execute_construct as execute_construct_graphdb
from .mustrdGraphDb import execute_select as execute_select_graphdb
from .spec_component import AnzoWhenSpec, WhenSpec
import logging

log = logging.getLogger(__name__)


def dispatch_upload_given(triple_store: dict, given: Graph):
    ts = triple_store['type']
    log.info(f"dispatch_upload_given to {ts}")
    return ts


upload_given = MultiMethod('upload_given', dispatch_upload_given)


@upload_given.method(TRIPLESTORE.RdfLib)
def _upload_given_rdflib(triple_store: dict, given: Graph):
    triple_store["given"] = given


@upload_given.method(TRIPLESTORE.GraphDb)
def _upload_given_graphdb(triple_store: dict, given: Graph):
    upload_given_graphdb(triple_store, given)


@upload_given.method(TRIPLESTORE.Anzo)
def _upload_given_anzo(triple_store: dict, given: Graph):
    upload_given_anzo(triple_store, given)


def dispatch_run_when(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    ts = triple_store['type']
    query_type = when.queryType
    log.info(f"dispatch_run_when to SPARQL type {query_type} to {ts}")
    return ts, query_type


run_when = MultiMethod('run_when', dispatch_run_when)


@run_when.method((TRIPLESTORE.Anzo, MUST.UpdateSparql))
def _anzo_run_when_update(spec_uri: URIRef, triple_store: dict, when: AnzoWhenSpec):
    log.debug(f"_anzo_run_when_update {spec_uri} {triple_store} {when} {type(when)}")
    if when.value is None:
        # fetch the query from the query step on anzo
        query = get_query_from_step(triple_store=when.spec_component_details.mustrd_triple_store,
                                                    query_step_uri=when.query_step_uri)
    else: 
        # we must already have the query
        query = when.value
    log.debug(f"_anzo_run_when_update.query {query}")
    return execute_update_anzo(triple_store, query, when.bindings)


@run_when.method((TRIPLESTORE.Anzo, MUST.ConstructSparql))
def _anzo_run_when_construct(spec_uri: URIRef, triple_store: dict, when: AnzoWhenSpec):
    return execute_construct_anzo(triple_store, when.value, when.bindings)


@run_when.method((TRIPLESTORE.Anzo, MUST.SelectSparql))
def _anzo_run_when_select(spec_uri: URIRef, triple_store: dict, when: AnzoWhenSpec):
    return execute_select_anzo(triple_store, when.value, when.bindings)


@run_when.method((TRIPLESTORE.GraphDb, MUST.UpdateSparql))
def _graphdb_run_when_update(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_update_graphdb(triple_store, when.value, when.bindings)


@run_when.method((TRIPLESTORE.GraphDb, MUST.ConstructSparql))
def _graphdb_run_when_construct(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_construct_graphdb(triple_store, when.value, when.bindings)


@run_when.method((TRIPLESTORE.GraphDb, MUST.SelectSparql))
def _graphdb_run_when_select(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_select_graphdb(triple_store, when.value, when.bindings)


@run_when.method((TRIPLESTORE.RdfLib, MUST.UpdateSparql))
def _rdflib_run_when_update(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_update_rdflib(triple_store, triple_store["given"], when.value, when.bindings)


@run_when.method((TRIPLESTORE.RdfLib, MUST.ConstructSparql))
def _rdflib_run_when_construct(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_construct_rdflib(triple_store, triple_store["given"], when.value, when.bindings)


@run_when.method((TRIPLESTORE.RdfLib, MUST.SelectSparql))
def _rdflib_run_when_select(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    return execute_select_rdflib(triple_store, triple_store["given"], when.value, when.bindings)


@run_when.method((TRIPLESTORE.Anzo, MUST.AnzoQueryDrivenUpdateSparql))
def _multi_run_when_anzo_query_driven_update(spec_uri: URIRef, triple_store: dict, when: AnzoWhenSpec):
    # run the parameters query to obtain the values for the template step and put them into a dictionary
    query_parameters = json.loads(execute_select_anzo(triple_store, when.paramQuery, None))
    if len(query_parameters['results']['bindings']) > 0:
        # replace the anzo query placeholders with the input and output graphs
        when_template = when.queryTemplate.replace(
            "${usingSources}",
            f"USING <{triple_store['input_graph']}> \nUSING <{triple_store['output_graph']}>").replace(
            "${targetGraph}", f"<{triple_store['output_graph']}>")

        # for each set of parameters insert their values into the template an run it
        for params in query_parameters['results']['bindings']:
            when_query = when_template
            for param in params:
                if params[param].get('datatype'):
                    value = params[param]['value']
                else:
                    if params[param]['type'] == 'uri':
                        value = '<' + params[param]['value'] + '>'
                    else:
                        value = '"' + params[param]['value'] + '"'
                when_query = when_query.replace("${" + param + "}", value)
            result = execute_update_anzo(triple_store, when_query, None)
        return result


@run_when.method(Default)
def _multi_run_when_default(spec_uri: URIRef, triple_store: dict, when: WhenSpec):
    log.error(f"run_when not implemented for {spec_uri} {triple_store} {when}")
    if when.queryType == MUST.AskSparql:
        log.warning(f"Skipping {spec_uri}, SPARQL ASK not implemented.")
        msg = "SPARQL ASK not implemented."
    elif when.queryType == MUST.DescribeSparql:
        log.warning(f"Skipping {spec_uri}, SPARQL DESCRIBE not implemented.")
        msg = "SPARQL DESCRIBE not implemented."
    elif triple_store['type'] not in [TRIPLESTORE.Anzo, TRIPLESTORE.GraphDb, TRIPLESTORE.RdfLib]:
        msg = f"{when.queryType} not implemented for {triple_store['type']}"
    else:
        log.warning(f"Skipping {spec_uri},  {when.queryType} is not a valid SPARQL query type.")
        msg = f"{when.queryType} is not a valid SPARQL query type."
    raise NotImplementedError(msg)



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
from typing import List
from urllib.parse import quote
from rdflib import Graph
import requests
from requests import Response, HTTPError, RequestException
from bs4 import BeautifulSoup
import logging

logger = logging.getLogger()


def query_azg(anzo_config: dict, query: str,
              format: str = "json", is_update: bool = False,
              data_layers: List[str] = None):
    params = {
        'skipCache': 'true',
        'format': format,
        'datasourceURI': anzo_config['gqe_uri'],
        'using-graph-uri' if is_update else 'default-graph-uri': data_layers,
        'using-named-graph-uri' if is_update else 'named-graph-uri': data_layers
    }
    url = f"{anzo_config['url']}/sparql"
    return send_anzo_query(anzo_config, url=url, params=params, query=query, is_update=is_update)


def query_graphmart(anzo_config: dict,
                    graphmart: str,
                    query: str,
                    format: str = "json",
                    data_layers: List[str] = None):
    params = {
        'skipCache': 'true',
        'format': format,
        'default-graph-uri': data_layers,
        'named-graph-uri': data_layers
    }

    url = f"{anzo_config['url']}/sparql/graphmart/{quote(graphmart, safe='')}"
    return send_anzo_query(anzo_config, url=url, params=params, query=query)


def query_configuration(anzo_config: dict, query: str, format: str = "json"):
    params = {
        'format': format,
        'includeMetadataGraphs': True
    }
    url = f"{anzo_config['url']}/sparql"
    return send_anzo_query(anzo_config, url=url, params=params, query=query)


# https://github.com/Semantic-partners/mustrd/issues/73
def manage_anzo_response(response: Response) -> str:
    content_string = response.content.decode("utf-8")
    if response.status_code == 200:
        logging.debug(f"Response content: {content_string}")
        return content_string
    elif response.status_code == 403:
        html = BeautifulSoup(content_string, 'html.parser')
        title_tag = html.title.string
        raise HTTPError(f"Anzo authentication error, status code: {response.status_code}, content: {title_tag}")
    else:
        raise RequestException(f"Anzo error, status code: {response.status_code}, content: {content_string}")


def send_anzo_query(anzo_config, url, params, query, is_update=False):
    headers = {"Content-Type": f"application/sparql-{'update' if is_update else 'query' }"}
    logger.debug(f"send_anzo_query {url=} {query=} {is_update=}")
    return manage_anzo_response(requests.post(url=url, params=params, data=query.encode('utf-8'),
                                              auth=(anzo_config['username'], anzo_config['password']),
                                              headers=headers, verify=False))


def json_to_dictlist(json_string: str):
    return list(map(lambda result: process_result(result), json.loads(json_string)['results']['bindings']))


def ttl_to_graph(ttl_string: str):
    return Graph().parse(data=ttl_string)


# Convert result to the right type
def process_result(result):
    xsd = "http://www.w3.org/2001/XMLSchema#"
    types = {
        f"{xsd}int": int,
        f"{xsd}integer": int,
        f"{xsd}decimal": float,
        f"{xsd}float": float,
        f"{xsd}double": float,
        f"{xsd}long": int
    }
    res = {}
    for key, type_value in result.items():
        if type_value['type'] in types:
            value = types[type_value['type']](type_value['value'])
        else:
            value = type_value['value']
        res.update({key: value})
    return res



import logging
from dataclasses import dataclass
import pytest
import os
from pathlib import Path, PosixPath
from rdflib.namespace import Namespace
from rdflib import Graph, RDF
from pytest import Session

from mustrd import logger_setup
from mustrd.TestResult import ResultList, TestResult, get_result_list
from mustrd.utils import get_mustrd_root
from mustrd.mustrd import (
    write_result_diff_to_log,
    get_triple_store_graph,
    get_triple_stores,
)
from mustrd.mustrd import (
    Specification,
    SpecSkipped,
    validate_specs,
    get_specs,
    SpecPassed,
    run_spec,
)
from mustrd.namespace import MUST, TRIPLESTORE, MUSTRDTEST
from typing import Union
from pyshacl import validate

import pathlib
import traceback

spnamespace = Namespace("https://semanticpartners.com/data/test/")

mustrd_root = get_mustrd_root()

MUSTRD_PYTEST_PATH = "mustrd_tests/"


def pytest_addoption(parser):
    group = parser.getgroup("mustrd option")
    group.addoption(
        "--mustrd",
        action="store_true",
        dest="mustrd",
        help="Activate/deactivate mustrd test generation.",
    )
    group.addoption(
        "--md",
        action="store",
        dest="mdpath",
        metavar="pathToMdSummary",
        default=None,
        help="create md summary file at that path.",
    )
    group.addoption(
        "--config",
        action="store",
        dest="configpath",
        metavar="pathToTestConfig",
        default=None,
        help="Ttl file containing the list of test to construct.",
    )
    group.addoption(
        "--secrets",
        action="store",
        dest="secrets",
        metavar="Secrets",
        default=None,
        help="Give the secrets by command line in order to be able to store secrets safely in CI tools",
    )
    group.addoption(
        "--pytest-path",
        action="store",
        dest="pytest_path",
        metavar="PytestPath",
        default=None,
        help="Filter tests based on the pytest_path property in .mustrd.ttl files.",
    )
    return


def pytest_configure(config) -> None:
    # Read configuration file
    if config.getoption("mustrd") and config.getoption("configpath"):
        config.pluginmanager.register(
            MustrdTestPlugin(
                config.getoption("mdpath"),
                Path(config.getoption("configpath")),
                config.getoption("secrets"),
            )
        )


def parse_config(config_path):
    test_configs = []
    config_graph = Graph().parse(config_path)
    shacl_graph = Graph().parse(
        Path(os.path.join(mustrd_root, "model/mustrdTestShapes.ttl"))
    )
    ont_graph = Graph().parse(
        Path(os.path.join(mustrd_root, "model/mustrdTestOntology.ttl"))
    )
    conforms, results_graph, results_text = validate(
        data_graph=config_graph,
        shacl_graph=shacl_graph,
        ont_graph=ont_graph,
        advanced=True,
        inference="none",
    )
    if not conforms:
        raise ValueError(
            f"Mustrd test configuration not conform to the shapes. SHACL report: {results_text}",
            results_graph,
        )

    for test_config_subject in config_graph.subjects(
        predicate=RDF.type, object=MUSTRDTEST.MustrdTest
    ):
        spec_path = get_config_param(
            config_graph, test_config_subject, MUSTRDTEST.hasSpecPath, str
        )
        data_path = get_config_param(
            config_graph, test_config_subject, MUSTRDTEST.hasDataPath, str
        )
        triplestore_spec_path = get_config_param(
            config_graph, test_config_subject, MUSTRDTEST.triplestoreSpecPath, str
        )
        pytest_path = get_config_param(
            config_graph, test_config_subject, MUSTRDTEST.hasPytestPath, str
        )
        filter_on_tripleStore = tuple(
            config_graph.objects(
                subject=test_config_subject, predicate=MUSTRDTEST.filterOnTripleStore
            )
        )

        # Root path is the mustrd test config path
        root_path = Path(config_path).parent
        spec_path = root_path / Path(spec_path) if spec_path else None
        data_path = root_path / Path(data_path) if data_path else None
        triplestore_spec_path = (
            root_path / Path(triplestore_spec_path) if triplestore_spec_path else None
        )

        test_configs.append(
            TestConfig(
                spec_path=spec_path,
                data_path=data_path,
                triplestore_spec_path=triplestore_spec_path,
                pytest_path=pytest_path,
                filter_on_tripleStore=filter_on_tripleStore,
            )
        )
    return test_configs


def get_config_param(config_graph, config_subject, config_param, convert_function):
    raw_value = config_graph.value(
        subject=config_subject, predicate=config_param, any=True
    )
    return convert_function(raw_value) if raw_value else None


@dataclass(frozen=True)
class TestConfig:
    spec_path: Path
    data_path: Path
    triplestore_spec_path: Path
    pytest_path: str
    filter_on_tripleStore: str = None


@dataclass(frozen=True)
class TestParamWrapper:
    id: str
    test_config: TestConfig
    unit_test: Union[Specification, SpecSkipped]


# Configure logging
logger = logger_setup.setup_logger(__name__)


class MustrdTestPlugin:
    md_path: str
    test_config_file: Path
    selected_tests: list
    secrets: str
    unit_tests: Union[Specification, SpecSkipped]
    items: list
    path_filter: str
    collect_error: BaseException

    def __init__(self, md_path, test_config_file, secrets):
        self.md_path = md_path
        self.test_config_file = test_config_file
        self.secrets = secrets
        self.items = []

    @pytest.hookimpl(tryfirst=True)
    def pytest_collection(self, session):
        logger.info("Starting test collection")
        
        self.unit_tests = []
        args = session.config.args
        
        # Split args into mustrd and regular pytest args
        mustrd_args = [arg for arg in args if ".mustrd.ttl" in arg]
        pytest_args = [arg for arg in args if arg != os.getcwd() and ".mustrd.ttl" not in arg]
        
        self.selected_tests = list(
            map(
                lambda arg: Path(arg.split("::")[0]),
                mustrd_args 
            )
        )
        logger.info(f"selected_tests is: {self.selected_tests}")

        self.path_filter = session.config.getoption("pytest_path") or None

        logger.info(f"path_filter is: {self.path_filter}")
        logger.info(f"Args: {args}")
        logger.info(f"Mustrd Args: {mustrd_args}")
        logger.info(f"Pytest Args: {pytest_args}")
        logger.info(f"Path Filter: {self.path_filter}")

        # Only modify args if we have mustrd tests to run
        if self.selected_tests:
            # Keep original pytest args and add config file for mustrd
            session.config.args = pytest_args + [str(self.test_config_file.resolve())]
        else:
            # Keep original args unchanged for regular pytest
            session.config.args = args

        logger.info(f"Final session.config.args: {session.config.args}")

    def get_file_name_from_arg(self, arg):
        if arg and len(arg) > 0 and "[" in arg and ".mustrd.ttl " in arg:
            return arg[arg.index("[") + 1 : arg.index(".mustrd.ttl ")]
        return None

    @pytest.hookimpl
    def pytest_collect_file(self, parent, path):
        logger.debug(f"Collecting file: {path}")
        # Only collect .ttl files that are mustrd suite config files
        if not str(path).endswith('.ttl'):
            return None
        if Path(path).resolve() != Path(self.test_config_file).resolve():
                logger.debug(f"{self.test_config_file}: Skipping non-matching-config file: {path}")
                return None
        
        mustrd_file = MustrdFile.from_parent(parent, path=pathlib.Path(path), mustrd_plugin=self)
        mustrd_file.mustrd_plugin = self
        return mustrd_file

    # Generate test for each triple store available
    def generate_tests_for_config(self, config, triple_stores, file_name):
        logger.debug(f"generate_tests_for_config {config=} {self=} {dir(self)}")
        shacl_graph = Graph().parse(
            Path(os.path.join(mustrd_root, "model/mustrdShapes.ttl"))
        )
        ont_graph = Graph().parse(Path(os.path.join(mustrd_root, "model/ontology.ttl")))
        logger.debug("Generating tests for config: " + str(config))
        logger.debug(f"selected_tests {self.selected_tests}")

        valid_spec_uris, spec_graph, invalid_spec_results = validate_specs(
            config,
            triple_stores,
            shacl_graph,
            ont_graph,
            file_name or "*",
            selected_test_files=self.selected_tests,
        )
        # Convert invalid specs to SpecInvalid instead of SpecSkipped
        invalid_specs = [
            SpecInvalid(
                spec.spec_uri,
                spec.triple_store,
                spec.message,
                spec.spec_file_name,
                spec.spec_source_file
            ) for spec in invalid_spec_results
        ]


        specs, skipped_spec_results = get_specs(
            valid_spec_uris, spec_graph, triple_stores, config
        )

        # Return normal specs + skipped results
        return specs + skipped_spec_results + invalid_specs

    # Function called to generate the name of the test
    def get_test_name(self, spec):
        # FIXME: SpecSkipped should have the same structure?
        if isinstance(spec, SpecSkipped):
            triple_store = spec.triple_store
        else:
            triple_store = spec.triple_store["type"]

        triple_store_name = triple_store.replace("https://mustrd.com/model/", "")
        test_name = spec.spec_uri.replace(spnamespace, "").replace("_", " ")
        return spec.spec_file_name + " : " + triple_store_name + ": " + test_name

    # Get triple store configuration or default
    def get_triple_stores_from_file(self, test_config):
        if test_config.triplestore_spec_path:
            try:
                triple_stores = get_triple_stores(
                    get_triple_store_graph(
                        test_config.triplestore_spec_path, self.secrets
                    )
                )
            except Exception as e:
                print(
                    f"""Triplestore configuration parsing failed {test_config.triplestore_spec_path}.
                    Only rdflib will be executed""",
                    e,
                )
                triple_stores = [
                    {"type": TRIPLESTORE.RdfLib, "uri": TRIPLESTORE.RdfLib}
                ]
        else:
            print("No triple store configuration required: using embedded rdflib")
            triple_stores = [{"type": TRIPLESTORE.RdfLib, "uri": TRIPLESTORE.RdfLib}]

        if test_config.filter_on_tripleStore:
            triple_stores = list(
                filter(
                    lambda triple_store: (
                        triple_store["uri"] in test_config.filter_on_tripleStore
                    ),
                    triple_stores,
                )
            )
        return triple_stores

    # Hook function. Initialize the list of result in session
    def pytest_sessionstart(self, session):
        session.results = dict()

    # Hook function called each time a report is generated by a test
    # The report is added to a list in the session
    # so it can be used later in pytest_sessionfinish to generate the global report md file
    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        outcome = yield
        result = outcome.get_result()

        if result.when == "call":
            # Add the result of the test to the session
            item.session.results[item] = result

    # Take all the test results in session, parse them, split them in mustrd and standard pytest  and generate md file
    def pytest_sessionfinish(self, session: Session, exitstatus):
        # if md path has not been defined in argument, then do not generate md file
        if not self.md_path:
            return

        test_results = []
        for test_conf, result in session.results.items():
            # Case auto generated tests
            if test_conf.originalname != test_conf.name:
                module_name = test_conf.parent.name
                class_name = test_conf.originalname
                test_name = (
                    test_conf.name.replace(class_name, "")
                    .replace("[", "")
                    .replace("]", "")
                )
                is_mustrd = True
            # Case normal unit tests
            else:
                module_name = test_conf.parent.parent.name
                class_name = test_conf.parent.name
                test_name = test_conf.originalname
                is_mustrd = False

            test_results.append(
                TestResult(
                    test_name, class_name, module_name, result.outcome, is_mustrd
                )
            )

        result_list = ResultList(
            None,
            get_result_list(
                test_results,
                lambda result: result.type,
                lambda result: is_mustrd and result.test_name.split("@")[1],
            ),
            False,
        )

        md = result_list.render()
        with open(self.md_path, "w") as file:
            file.write(md)

@dataclass(frozen=True)
class SpecInvalid:
    spec_uri: str
    triple_store: str
    message: str
    spec_file_name: str = None
    spec_source_file: Path = None

class MustrdFile(pytest.File):
    mustrd_plugin: MustrdTestPlugin

    def __init__(self, *args, mustrd_plugin, **kwargs):
        logger.debug(f"Creating MustrdFile with args: {args}, kwargs: {kwargs}")
        self.mustrd_plugin = mustrd_plugin
        super(pytest.File, self).__init__(*args, **kwargs)

    def collect(self):
        try:
            logger.info(f"{self.mustrd_plugin.test_config_file}: Collecting tests from file: {self.path=}")
            # Only process the specific mustrd config file we were given
            
            # if not str(self.fspath).endswith(".ttl"):
            #     return []
            # Only process the specific mustrd config file we were given
            # if str(self.fspath) != str(self.mustrd_plugin.test_config_file):
            #     logger.info(f"Skipping non-config file: {self.fspath}")
            #     return []
                
            test_configs = parse_config(self.path)
            from collections import defaultdict
            pytest_path_grouped = defaultdict(list)
            for test_config in test_configs:
                if (
                    self.mustrd_plugin.path_filter is not None
                    and not str(test_config.pytest_path).startswith(str(self.mustrd_plugin.path_filter))
                ):
                    logger.info(f"Skipping test config due to path filter: {test_config.pytest_path=} {self.mustrd_plugin.path_filter=}")
                    continue
                
                triple_stores = self.mustrd_plugin.get_triple_stores_from_file(test_config)
                try:
                    specs = self.mustrd_plugin.generate_tests_for_config(
                        {
                            "spec_path": test_config.spec_path,
                            "data_path": test_config.data_path,
                        },
                        triple_stores,
                        None,
                    )
                except Exception as e:
                    logger.error(f"Error generating tests: {e}\n{traceback.format_exc()}")
                    specs = [
                        SpecInvalid(
                            MUST.TestSpec,
                            triple_store["uri"] if isinstance(triple_store, dict) else triple_store,
                            f"Test generation failed: {str(e)}",
                            spec_file_name=str(test_config.spec_path.name) if test_config.spec_path else "unknown.mustrd.ttl",
                            spec_source_file=self.path if test_config.spec_path else Path("unknown.mustrd.ttl"),
                        )
                        for triple_store in (triple_stores or test_config.filter_on_tripleStore)
                    ]
                pytest_path = getattr(test_config, "pytest_path", "unknown")
                for spec in specs:
                    pytest_path_grouped[pytest_path].append(spec)

            for pytest_path, specs_for_path in pytest_path_grouped.items():
                logger.info(f"pytest_path group: {pytest_path} ({len(specs_for_path)} specs)")

                yield MustrdPytestPathCollector.from_parent(
                    self,
                    name=str(pytest_path),
                    pytest_path=pytest_path,
                    specs=specs_for_path,
                    mustrd_plugin=self.mustrd_plugin,
                )
        except Exception as e:
            self.mustrd_plugin.collect_error = e
            logger.error(f"Error during collection {self.path}: {type(e)} {e} {traceback.format_exc()}")
            raise e


class MustrdPytestPathCollector(pytest.Class):
    def __init__(self, name, parent, pytest_path, specs, mustrd_plugin):
        super().__init__(name, parent)
        self.pytest_path = pytest_path
        self.specs = specs
        self.mustrd_plugin = mustrd_plugin

    def collect(self):
        for spec in self.specs:
            item = MustrdItem.from_parent(
                self,
                name=spec.spec_file_name,
                spec=spec,
            )
            self.mustrd_plugin.items.append(item)
            yield item


class MustrdItem(pytest.Item):
    def __init__(self, name, parent, spec):
        logging.debug(f"Creating item: {name}")
        super().__init__(name, parent)
        self.spec = spec
        self.fspath = spec.spec_source_file
        self.originalname = name

    def runtest(self):
        result = run_test_spec(self.spec)
        if not result:
            raise AssertionError(f"Test {self.name} failed")

    def repr_failure(self, excinfo):
        # excinfo.value is the exception instance
        # You can add more context here
        tb_lines = traceback.format_exception(excinfo.type, excinfo.value, excinfo.tb)
        tb_str = "".join(tb_lines)
        return (
            f"{self.name} failed:\n"
            f"Spec: {self.spec.spec_uri}\n"
            f"File: {self.spec.spec_source_file}\n"
            f"Error: \n{excinfo.value}\n"
            f"Traceback:\n{tb_str}"
        )
    
    def reportinfo(self):
        r = "", 0, f"mustrd test: {self.name}"
        return r


# Function called in the test to actually run it
def run_test_spec(test_spec):
    if isinstance(test_spec, SpecSkipped):
        pytest.skip(f"Invalid configuration, error : {test_spec.message}")
    result = run_spec(test_spec)
    result_type = type(result)
    if isinstance(test_spec, SpecInvalid):
        raise ValueError(f"Invalid test specification: {test_spec.message} {test_spec}")
    if result_type == SpecSkipped:
        # FIXME: Better exception management
        pytest.skip("Unsupported configuration")
    if result_type != SpecPassed:
        write_result_diff_to_log(result, logger.info)
        log_lines = []
        def log_to_string(message):
            log_lines.append(message)
        write_result_diff_to_log(result, log_to_string)
        raise AssertionError("Test failed: " + "\n".join(log_lines))

    return result_type == SpecPassed



"""
MIT License

Copyright (c) 2023 Semantic Partners Ltd

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from rdflib import Graph, ConjunctiveGraph, Literal, URIRef
from requests import ConnectTimeout, HTTPError, ConnectionError
import logging
from mustrd.anzo_utils import query_azg, query_graphmart
from mustrd.anzo_utils import query_configuration, json_to_dictlist, ttl_to_graph



def execute_select(triple_store: dict,  when: str, bindings: dict = None) -> str:
    try:
        if bindings:
            when = query_with_bindings(bindings, when)
        # FIXME: why do we have those tokens in a select query? in particular ${targetGraph}?
        # FIXME: why do we also query the output graph?
        when = when.replace("${fromSources}",
                            f"FROM <{triple_store['input_graph']}>\nFROM <{triple_store['output_graph']}>").replace(
                                "${targetGraph}", f"<{triple_store['output_graph']}>")
        # TODO: manage results here
        return query_azg(anzo_config=triple_store, query=when, data_layers=[triple_store['input_graph']])
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise


def execute_update(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    logging.debug(f"updating in anzo! {triple_store=} {when=}")
    input_graph = triple_store['input_graph']
    output_graph = triple_store['output_graph']

    # FIXME: that will only work with steps.
    # We could replace USING clauses with using-graph-uri parameter
    # But there is no parameter for default insert graphs.
    substituted_query = when.replace("${usingSources}",
                                     f"""USING <{input_graph}>
USING <{triple_store['output_graph']}>""").replace(
                                         "${targetGraph}", f"<{output_graph}>")

    response = query_azg(anzo_config=triple_store, query=substituted_query, is_update=True,
                         data_layers=[input_graph, output_graph], format="ttl")
    logging.debug(f'response {response}')
    # TODO: deal with error responses
    new_graph = ttl_to_graph(query_azg(anzo_config=triple_store, query="construct {?s ?p ?o} { ?s ?p ?o }",
                                       format="ttl", data_layers=output_graph))
    logging.debug(f"new_graph={new_graph.serialize(format='ttl')}")
    return new_graph


def execute_construct(triple_store: dict, when: str, bindings: dict = None) -> Graph:
    try:
        if bindings:
            when = query_with_bindings(bindings, when)
        return ttl_to_graph(query_azg(anzo_config=triple_store, query=when, format="ttl",
                                      data_layers=triple_store['input_graph']))
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout) as e:
        logging.error(f'response {e}')
        raise


def query_with_bindings(bindings: dict, when: str) -> str:
    values = ""
    for key, value in bindings.items():
        values += f"VALUES ?{key} {{{value.n3()}}} "
    split_query = when.lower().split("where {", 1)
    return f"{split_query[0].strip()} WHERE {{ {values} {split_query[1].strip()}"


# Get Given or then from the content of a graphmart
def get_spec_component_from_graphmart(triple_store: dict, graphmart: URIRef, layer: URIRef = None) -> ConjunctiveGraph:
    try:
        return ttl_to_graph(query_graphmart(anzo_config=triple_store, graphmart=graphmart,
                                            query="CONSTRUCT {?s ?p ?o} WHERE {?s ?p ?o}",
                                            data_layers=layer, format="ttl"))
    except RuntimeError as e:
        raise ConnectionError(f"Anzo connection error, {e}")


def get_query_from_querybuilder(triple_store: dict, folder_name: Literal, query_name: Literal) -> str:
    query = f"""SELECT ?query WHERE {{
        graph ?queryFolder {{
            ?bookmark a <http://www.cambridgesemantics.com/ontologies/QueryPlayground#QueryBookmark>;
                        <http://openanzo.org/ontologies/2008/07/System#query> ?query;
                        <http://purl.org/dc/elements/1.1/title> "{query_name}"
            }}
            ?queryFolder a <http://www.cambridgesemantics.com/ontologies/QueryPlayground#QueryFolder>;
                        <http://purl.org/dc/elements/1.1/title> "{folder_name}"
    }}"""
    result = json_to_dictlist(query_configuration(
        anzo_config=triple_store, query=query))
    if len(result) == 0:
        raise FileNotFoundError(
            f"Query {query_name} not found in folder {folder_name}")
    return result[0].get("query")


# https://github.com/Semantic-partners/mustrd/issues/102
def get_query_from_step(triple_store: dict, query_step_uri: URIRef) -> str:
    query = f"""SELECT ?query WHERE {{
        BIND(<{query_step_uri}> as ?stepUri)
            ?stepUri a <http://cambridgesemantics.com/ontologies/Graphmarts#Step>;
                     <http://cambridgesemantics.com/ontologies/Graphmarts#transformQuery> ?query
    }}"""
    return json_to_dictlist(query_configuration(anzo_config=triple_store, query=query))[0]['query']


def get_queries_from_templated_step(triple_store: dict, query_step_uri: URIRef) -> dict:
    query = f"""SELECT ?param_query ?query_template WHERE {{
        BIND(<{query_step_uri}> as ?stepUri)
            ?stepUri    a <http://cambridgesemantics.com/ontologies/Graphmarts#Step> ;
                        <http://cambridgesemantics.com/ontologies/Graphmarts#parametersTemplate> ?param_query ;
                        <http://cambridgesemantics.com/ontologies/Graphmarts#template> ?query_template .
    }}
    """
    return json_to_dictlist(query_configuration(anzo_config=triple_store, query=query))[0]


def get_queries_for_layer(triple_store: dict, graphmart_layer_uri: URIRef):
    query = f"""PREFIX graphmarts: <http://cambridgesemantics.com/ontologies/Graphmarts#>
    PREFIX anzo: <http://openanzo.org/ontologies/2008/07/Anzo#>
SELECT ?query ?param_query ?query_template
  {{ <{graphmart_layer_uri}> graphmarts:step ?step .
  ?step         anzo:index ?index ;
                anzo:orderedValue ?query_step .
  ?query_step graphmarts:enabled true ;
  OPTIONAL {{  ?query_step
                graphmarts:parametersTemplate ?param_query ;
                graphmarts:template ?query_template ;
      . }}
  OPTIONAL {{  ?query_step
                graphmarts:transformQuery ?query ;
      . }}
  }}
  ORDER BY ?index"""
    return json_to_dictlist(query_configuration(anzo_config=triple_store, query=query))


def upload_given(triple_store: dict, given: Graph):
    logging.debug(f"upload_given {triple_store} {given}")

    try:
        clear_graph(triple_store, triple_store['input_graph'])
        clear_graph(triple_store, triple_store['output_graph'])
        serialized_given = given.serialize(format="nt")
        insert_query = f"INSERT DATA {{graph <{triple_store['input_graph']}>{{{serialized_given}}}}}"
        query_azg(anzo_config=triple_store, query=insert_query, is_update=True)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise


def clear_graph(triple_store: dict, graph_uri: str):
    try:
        clear_query = f"CLEAR GRAPH <{graph_uri}>"
        query_azg(anzo_config=triple_store, query=clear_query, is_update=True)
    except (ConnectionError, TimeoutError, HTTPError, ConnectTimeout):
        raise
